import { doesExist, mustDefault } from '@apextoaster/js-utils';

import { LockData, LockType } from '../lock.js';
import { printLock } from '../utils/lock.js';
import { walkPath } from '../utils/path.js';
import { CommandContext } from './index.js';

export async function checkCommand(context: CommandContext) {
  const { args, logger } = context;
  const { now } = args;

  logger.info({ now }, 'running check command');

  const [allowed] = await checkArgsPath(context);
  return allowed;
}

export interface CheckOptions {
  path: string;
  type: LockType;
  now: number;
  recursive: boolean;
}

export function getCheckPaths(args: CheckOptions): Array<string> {
  if (args.recursive) {
    return walkPath(args.path);
  } else {
    return [args.path];
  }
}

/**
 * This is the core of the check logic, including the path recursion and allow type logic.
 */
export async function checkArgsPath(context: CommandContext, overrideArgs?: CheckOptions): Promise<[boolean, Array<LockData>]> {
  const { logger, storage } = context;
  const args = mustDefault(overrideArgs, context.args);

  const locks: Array<LockData> = [];
  const paths = getCheckPaths(args);
  for (const path of paths) {
    const lock = await storage.get(path);

    if (doesExist(lock)) {
      if (lock.expires_at < args.now) {
        logger.debug({ lock, path }, 'found expired lock');
      } else {
        const friendly = printLock(path, lock);
        logger.debug({ lock, friendly, path }, 'found active lock');
        locks.push(lock);

        if (lock.allow.includes(args.type)) {
          logger.warn({ lock, friendly, path }, 'found existing lock, check type is allowed');
        } else {
          logger.warn({ lock, friendly, path }, 'found existing lock, check type is not allowed');
        }
      }
    }
  }

  const allowed = locks.every((lock) => lock.allow.includes(args.type));
  return [allowed, locks];
}
