/* eslint-disable camelcase */
import { doesExist, mustDefault } from '@apextoaster/js-utils';

import { ParsedArgs } from '../args.js';
import { LOCK_TYPES, LockCI, LockData } from '../lock.js';
import { calculateExpires, printTime } from './time.js';

export const ENV_UNSET = 'not set';

export function buildCI(args: ParsedArgs, env: typeof process.env): LockCI | undefined {
  if (doesExist(env.GITLAB_CI)) {
    return {
      project: mustDefault(args['ci-project'], env.CI_PROJECT_PATH, ENV_UNSET),
      ref: mustDefault(args['ci-ref'], env.CI_COMMIT_REF_SLUG, ENV_UNSET),
      commit: mustDefault(args['ci-commit'], env.CI_COMMIT_SHA, ENV_UNSET),
      pipeline: mustDefault(args['ci-pipeline'], env.CI_PIPELINE_ID, ENV_UNSET),
      job: mustDefault(args['ci-job'], env.CI_JOB_ID, ENV_UNSET),
    };
  }

  return undefined;
}

export function buildAuthor(args: ParsedArgs, env: typeof process.env): string {
  if (doesExist(env.GITLAB_CI)) {
    return mustDefault(args.author, env.GITLAB_USER_EMAIL, env.USER);
  }

  return mustDefault(args.author, env.USER);
}

export function buildLinks(args: ParsedArgs): Record<string, string> {
  const links: Record<string, string> = {};

  for (const link of args.links) {
    const match = /^([-a-z]+):(.+)$/.exec(link);
    if (doesExist(match)) {
      const [_full, name, url] = Array.from(match);
      links[name] = url;
    }
  }

  return links;
}

export function buildLock(args: ParsedArgs, env = process.env): LockData {
  return {
    type: mustDefault(args.type, 'deploy'),
    path: args.path,
    author: buildAuthor(args, env),
    links: buildLinks(args),
    allow: args.allow,
    created_at: args.now,
    updated_at: args.now,
    expires_at: calculateExpires(args),
    source: args.source,
    ci: buildCI(args, env),
  };
}

export function printLinks(links: Record<string, string>): Array<string> {
  return Object.keys(links).map((name) => `${name}: ${links[name]}`);
}

export function printLock(path: string, data: LockData): string {
  const friendlyExpires = printTime(data.expires_at);
  const friendlyLinks = printLinks(data.links);
  const friendlyType = LOCK_TYPES[data.type];

  const base = `${path} is locked until ${friendlyExpires} by ${friendlyType} in ${data.source}`;
  return [
    base,
    ...friendlyLinks,
  ].join(', ');
}
