import { doesExist, InvalidArgumentError } from '@apextoaster/js-utils';

import { ParsedArgs } from '../args.js';

export const MILLIS_PER_SECOND = 1000;
export const SECONDS_PER_MINUTE = 60;
export const MINUTES_PER_HOUR = 60;
export const HOURS_PER_DAY = 24;
export const DAYS_PER_YEAR = 365;

export const TIME_MULTIPLIERS: Array<[RegExp, number]> = [
  [/^(\d+)$/, 1], // integer seconds
  [/^(\d+)s$/, 1], // human seconds
  [/^(\d+)m$/, SECONDS_PER_MINUTE], // human minutes
  [/^(\d+)h$/, SECONDS_PER_MINUTE * MINUTES_PER_HOUR], // human hours
  [/^(\d+)d$/, SECONDS_PER_MINUTE * MINUTES_PER_HOUR * HOURS_PER_DAY], // human days
  [/^(\d+)y$/, SECONDS_PER_MINUTE * MINUTES_PER_HOUR * HOURS_PER_DAY * DAYS_PER_YEAR], // human years
];

export const TIME_ISO_MATCH = /^\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}Z?/;

export function dateToSeconds(date: Date): number {
  return Math.round(date.getTime() / MILLIS_PER_SECOND);
}

/**
 * Convert a string of the form `12345` or `15m` into seconds.
 */
export function parseTime(time: number | string): number {
  if (typeof time === 'number') {
    return time;
  }

  if (TIME_ISO_MATCH.test(time)) {
    return dateToSeconds(new Date(time));
  }

  for (const [regex, mult] of TIME_MULTIPLIERS) {
    const match = time.match(regex);
    if (doesExist(match)) {
      const [_full, digits] = Array.from(match);
      return parseInt(digits, 10) * mult;
    }
  }

  throw new InvalidArgumentError('invalid time');
}

export function printTime(time: number): string {
  return new Date(time * MILLIS_PER_SECOND).toLocaleString();
}

export function calculateExpires(args: ParsedArgs): number {
  if (doesExist(args.until)) {
    return parseTime(args.until);
  }

  if (doesExist(args.duration)) {
    return args.now + parseTime(args.duration);
  }

  throw new InvalidArgumentError('must provide either duration or until');
}
